@@ -23,16 +23,18 @@
 # SPDX-License-Identifier: LGPL-2.0-or-later
 
 from __future__ import annotations
-from typing_extensions import Self
 
 import sys
 import inspect
 import itertools
 import weakref
 import typing
+import gi
 
-from gi.overrides import override
+gi.require_version('GLib', '2.0')
+gi.require_version('GObject', '2.0')
 from gi.repository import GLib, GObject
+from gi.overrides import override
 
 # Typing relies on https://github.com/pygobject/pygobject-stubs.
 if typing.TYPE_CHECKING:
@@ -44,6 +46,7 @@ if typing.TYPE_CHECKING:
     # Bin.add(). The type checker will use signature from stubs which is our
     # override signature.
     from gi.repository import Gst
+    from typing_extensions import Self
 
     # Type annotations cannot have `Gst.` prefix because they are copied into
     # Gst stubs module which cannot refer to itself. Use type aliases.
@@ -74,6 +77,17 @@ python module to use with Gst 0.10"
     warnings.warn(warn_msg, RuntimeWarning)
 
 
+class Float(float):
+    '''
+    A wrapper to force conversion to G_TYPE_FLOAT instead of G_TYPE_DOUBLE when
+    used in e.g. Gst.ValueArray.
+    '''
+    __gtype__ = GObject.TYPE_FLOAT
+
+
+__all__.append('Float')
+
+
 # Ensuring that PyGObject loads the URIHandler interface
 # so we can force our own implementation soon enough (in gstmodule.c)
 class URIHandler(Gst.URIHandler):
@@ -111,12 +125,9 @@ class Bin(Gst.Bin):
     def make_and_add(self, factoryname: str, name: typing.Optional[str] = None) -> Element:
         '''
         :raises Gst.AddError:
+        :raises Gst.MissingPluginError:
         '''
-        elem = Gst.ElementFactory.make(factoryname, name)
-        if not elem:
-            raise AddError(
-                'No such element: {}'.format(factoryname))
-        assert isinstance(elem, Element)  # Tell mypy we actually have our override subclass
+        elem = ElementFactory.make(factoryname, name)
         self.add(elem)
         return elem
 
@@ -233,31 +244,25 @@ class Caps(MiniObjectMixin, Gst.Caps):  # type: ignore[misc]
     def __nonzero__(self):
         return not self.is_empty()
 
-    def __new__(cls, *args):
-        if not args:
-            return Caps.new_empty()
-        if len(args) > 1:
-            raise TypeError("wrong arguments when creating GstCaps object")
-
-        assert len(args) == 1
-        if isinstance(args[0], str):
-            return Caps.from_string(args[0])
-        elif isinstance(args[0], Caps):
-            return args[0].copy()
-        elif isinstance(args[0], Structure):
+    @staticmethod
+    def __new__(cls: type[Self], arg: typing.Optional[typing.Union[str, Caps, Structure, list[Structure], tuple[Structure, ...]]] = None) -> Self:
+        if not arg:
+            return Caps.new_empty()  # type: ignore[return-value]
+        elif isinstance(arg, str):
+            return Caps.from_string(arg)  # type: ignore[return-value]
+        elif isinstance(arg, Caps):
+            return arg.copy()  # type: ignore[return-value]
+        elif isinstance(arg, Structure):
             res = Caps.new_empty()
-            res.append_structure(args[0])
-            return res
-        elif isinstance(args[0], (list, tuple)):
+            res.append_structure(arg)
+            return res  # type: ignore[return-value]
+        elif isinstance(arg, (list, tuple)):
             res = Caps.new_empty()
-            for e in args[0]:
+            for e in arg:
                 res.append_structure(e)
-            return res
+            return res  # type: ignore[return-value]
 
-        raise TypeError("wrong arguments when creating GstCaps object")
-
-    def __init__(self, *args, **kwargs):
-        return super(Caps, self).__init__()
+        raise TypeError(f"wrong arguments when creating GstCaps object")
 
     def __str__(self) -> str:
         return self.to_string()
@@ -265,6 +270,10 @@ class Caps(MiniObjectMixin, Gst.Caps):  # type: ignore[misc]
     def __getitem__(self, index: int) -> Structure:
         return self.get_structure(index)
 
+    def __iter__(self) -> typing.Iterator[Structure]:
+        for i in range(self.get_size()):
+            yield self.get_structure(i)
+
     def __len__(self) -> int:
         return self.get_size()
 
@@ -407,6 +416,13 @@ class IteratorError(Exception):
 __all__.append('IteratorError')
 
 
+class MissingPluginError(Exception):
+    pass
+
+
+__all__.append('MissingPluginError')
+
+
 class AddError(Exception):
     pass
 
@@ -458,11 +474,13 @@ class ElementFactory(Gst.ElementFactory):
         return self.get_metadata("klass")
 
     @staticmethod
-    def make(factoryname: str, name: typing.Optional[str] = None) -> typing.Optional[Element]:  # type: ignore[override]
+    def make(factoryname: str, name: typing.Optional[str] = None) -> Element:  # type: ignore[override]
         '''
         :raises Gst.PluginMissingError:
         '''
         elem = Gst.ElementFactory.make(factoryname, name)
+        if not elem:
+            raise MissingPluginError(f'No such element: {factoryname}')
         return elem  # type: ignore[return-value]
 
 
@@ -494,52 +512,60 @@ __all__.append('StructureContextManager')
 
 
 class Structure(Gst.Structure):
-    def __new__(cls, *args, **kwargs):
-        if not args:
-            if kwargs:
-                raise TypeError("wrong arguments when creating GstStructure, first argument"
-                                " must be the structure name.")
-            struct = Structure.new_empty()
-            return struct
-        elif len(args) > 1:
-            raise TypeError("wrong arguments when creating GstStructure object")
-        elif isinstance(args[0], str):
+    @staticmethod
+    def __new__(cls: type[Self], arg: typing.Union[str, Structure], **kwargs) -> Self:
+        if isinstance(arg, str):
             if not kwargs:
-                struct = Structure.from_string(args[0])[0]
-                return struct
-            struct = Structure.new_empty(args[0])
+                struct = Structure.from_string(arg)[0]
+                return struct  # type: ignore[return-value]
+            struct = Structure.new_empty(arg)
             for k, v in kwargs.items():
                 struct[k] = v
-
-            return struct
-        elif isinstance(args[0], Structure):
-            struct = args[0].copy()
-            return struct
+            return struct  # type: ignore[return-value]
+        elif isinstance(arg, Structure):
+            struct = arg.copy()
+            return struct  # type: ignore[return-value]
 
         raise TypeError("wrong arguments when creating GstStructure object")
 
-    def __init__(self, *args, **kwargs):
-        pass
-
     def __ptr__(self):
         return _gi_gst._get_object_ptr(self)
 
     def __getitem__(self, key: str) -> typing.Any:
-        return self.get_value(key)
-
-    def keys(self) -> typing.Iterable[str]:
-        keys = set()
-
-        def foreach(fid, value, unused1, udata):
-            keys.add(GLib.quark_to_string(fid))
-            return True
-
-        self.foreach(foreach, None, None)
-        return keys
+        val = self.get_value(key)
+        if val is None:
+            raise KeyError(f"key {key} not found")
+        return val
 
     def __setitem__(self, key: str, value: typing.Any) -> None:
         self.set_value(key, value)
 
+    def __len__(self) -> int:
+        return self.n_fields()
+
+    def __iter__(self) -> typing.Iterator[str]:
+        return self.keys()
+
+    def items(self) -> typing.Iterator[typing.Tuple[str, typing.Any]]:
+        pairs: typing.List[typing.Tuple[str, typing.Any]] = []
+
+        def foreach(fid, value):
+            pairs.append((GLib.quark_to_string(fid), value))
+            return True
+
+        self.foreach(foreach)
+        return iter(pairs)
+
+    def keys(self) -> typing.Iterator[str]:
+        keys: list[str] = []
+
+        def foreach(fid, value):
+            keys.append(GLib.quark_to_string(fid))
+            return True
+
+        self.foreach(foreach)
+        return iter(keys)
+
     def set_value(self, key: str, value: typing.Any) -> bool:
         if not _gi_gst.structure_is_writable(self):
             raise NotWritableStructure("Trying to write to a not writable structure."
@@ -727,6 +753,8 @@ class Int64Range(Gst.Int64Range):
 
 
 class Bitmask(Gst.Bitmask):
+    v: int
+
     def __init__(self, v: int) -> None:
         if not isinstance(v, int):
             raise TypeError(f"{type(v)} is not an int.")
@@ -749,7 +777,10 @@ __all__.append('Int64Range')
 
 
 class DoubleRange(Gst.DoubleRange):
-    def __init__(self, start: int | float, stop: int | float):
+    start: float
+    stop: float
+
+    def __init__(self, start: float, stop: float):
         self.start = float(start)
         self.stop = float(stop)
 
@@ -768,6 +799,9 @@ __all__.append('DoubleRange')
 
 
 class FractionRange(Gst.FractionRange):
+    start: Fraction
+    stop: Fraction
+
     def __init__(self, start: Fraction, stop: Fraction):
         if not isinstance(start, Fraction):
             raise TypeError(f"{type(start)} is not a Gst.Fraction.")
@@ -793,8 +827,29 @@ __all__.append('FractionRange')
 
 
 class ValueArray(Gst.ValueArray):
-    def __init__(self, array: typing.List[typing.Any]):
-        self.array = list(array)
+    def __init__(self, array: typing.Optional[typing.List[typing.Any]] = None):
+        self.array = list(array or [])
+
+    def append(self, item: typing.Any) -> None:
+        self.array.append(item)
+
+    def prepend(self, item: typing.Any) -> None:
+        self.array = [item] + self.array
+
+    @staticmethod
+    def append_value(this: ValueArray, item: typing.Any) -> None:
+        this.append(item)
+
+    @staticmethod
+    def prepend_value(this: ValueArray, item: typing.Any) -> None:
+        this.prepend(item)
+
+    @staticmethod
+    def get_size(this: ValueArray) -> int:
+        return len(this.array)
+
+    def __iter__(self) -> typing.Iterator[typing.Any]:
+        return iter(self.array)
 
     def __getitem__(self, index: int) -> typing.Any:
         return self.array[index]
@@ -817,8 +872,29 @@ __all__.append('ValueArray')
 
 
 class ValueList(Gst.ValueList):
-    def __init__(self, array: typing.List[typing.Any]):
-        self.array = list(array)
+    def __init__(self, array: typing.Optional[typing.List[typing.Any]] = None):
+        self.array = list(array or [])
+
+    def append(self, item: typing.Any) -> None:
+        self.array.append(item)
+
+    def prepend(self, item: typing.Any) -> None:
+        self.array = [item] + self.array
+
+    @staticmethod
+    def append_value(this: ValueList, item: typing.Any) -> None:
+        this.append(item)
+
+    @staticmethod
+    def prepend_value(this: ValueList, item: typing.Any) -> None:
+        this.prepend(item)
+
+    @staticmethod
+    def get_size(this: ValueList) -> int:
+        return len(this.array)
+
+    def __iter__(self) -> typing.Iterator[typing.Any]:
+        return iter(self.array)
 
     def __getitem__(self, index: int) -> typing.Any:
         return self.array[index]
@@ -893,13 +969,18 @@ def pairwise(iterable: typing.Iterable[Element]) -> typing.Iterator[tuple[Elemen
 
 
 class MapInfo:
+    memory: typing.Optional[Memory]
+    flags: MapFlags
+    size: int
+    maxsize: int
+    data: typing.Optional[memoryview]
+
     def __init__(self):
         self.memory = None
         self.flags = Gst.MapFlags(0)
         self.size = 0
         self.maxsize = 0
         self.data = None
-        self.user_data = None
         self.__parent__ = None
 
     def __iter__(self):
@@ -918,6 +999,9 @@ class MapInfo:
         if not self.__parent__.unmap(self):
             raise MapError('MappingError', 'Unmapping was not successful')
 
+    def get_data(self):
+        return self.data
+
 
 __all__.append("MapInfo")
 
@@ -1052,14 +1136,48 @@ def fake_method(*args):
     raise NotInitialized("Please call Gst.init(argv) before using GStreamer")
 
 
+def find_gi_repository_parent(klass):
+    """Find the gi.repository parent class in the MRO for introspection methods"""
+    for parent in klass.__mro__:
+        if parent.__module__.startswith('gi.repository.'):
+            return parent
+    return None
+
+
+def _collect_methods_from_dict(source_dict, klass, seen):
+    """Helper to collect methods from a class dictionary, avoiding dunder methods."""
+    methods = []
+    for attr_name in source_dict:
+        if attr_name in seen:
+            continue
+        attr = source_dict[attr_name]
+        if isinstance(attr, (type(Gst.init), staticmethod, classmethod)):
+            # Skip dunder methods as they're Python special methods used for
+            # object instantiation and other internals, replacing them breaks
+            # class behavior
+            if not attr_name.startswith('__'):
+                methods.append((attr_name, getattr(klass, attr_name)))
+            seen.add(attr_name)
+    return methods
+
+
 real_functions = [o for o in inspect.getmembers(Gst) if isinstance(o[1], type(Gst.init))]
 
 class_methods = []
 for cname_klass in [o for o in inspect.getmembers(Gst) if isinstance(o[1], type(Gst.Element)) or isinstance(o[1], type(Gst.Caps))]:
-    class_methods.append((cname_klass,
-                         [(o, cname_klass[1].__dict__[o])
-                          for o in cname_klass[1].__dict__
-                          if isinstance(cname_klass[1].__dict__[o], type(Gst.init))]))
+    klass = cname_klass[1]
+    methods = []
+    seen: set[str] = set()
+
+    # Collect methods from the override class itself
+    methods.extend(_collect_methods_from_dict(klass.__dict__, klass, seen))
+
+    # Collect methods from the gi.repository introspection parent class
+    gi_parent = find_gi_repository_parent(klass)
+    if gi_parent:
+        methods.extend(_collect_methods_from_dict(gi_parent.__dict__, klass, seen))
+
+    class_methods.append((cname_klass, methods))
 
 pre_init_functions = set([
     "init",
@@ -1107,7 +1225,11 @@ def init(argv: typing.Optional[list[str]] = None) -> None:
     if Gst.is_initialized():
         return
 
-    return real_init(argv)
+    # FIXME: Workaround for pygobject handling nullability wrong
+    if argv is None:
+        argv = []
+
+    real_init(argv)
 
 
 Gst.init = init
